import warnings
from typing import Optional
from ._types import UserChannel, UserClientType, GameType
from ._exceptions import ComboTokenInvalidError


class User:
    """
    用户类。
    """

    def __init__(
            self,
            combo_token: str,
            sys_version: str,
            device_id: str,
            device_name: str,
            device_model: str,
            client_type: UserClientType,
            *,
            game_type: Optional[GameType] = None,
            channel: Optional[UserChannel] = UserChannel.Official
    ):
        """
        创建一个用户。

        Args:
            combo_token (str): 对应 headers 中的 x-rpc-combo_token。
            sys_version (str): 对应 headers 中的 x-rpc-sys_version。
            device_id (str): 对应 headers 中的 x-rpc-device_id。
            device_name (str): 对应 headers 中的 x-rpc-device_name。
            device_model (str): 对应 headers 中的 x-rpc-device_model。
            client_type (UserClientType): 用户的客户端种类。
            game_type (Optional[GameType]): 游戏类型，若为空则将会从 combo_token 中自动识别。
            channel (Optional[UserChannel]): 用户的游戏渠道。
        """
        self._combo_token = combo_token
        self._sys_version = sys_version
        self._device_id = device_id
        self._device_name = device_name
        self._device_model = device_model
        self._client_type = client_type
        self._game_type = game_type
        self._channel = channel

        # Automatic detection of the game type.

        try:
            bi = self._combo_token.split(";bi=")[1]
        except IndexError:
            raise ComboTokenInvalidError(
                "An error occurred in the automatic detection of the game type, "
                "the 'bi' segment was not found in combo token."
            )

        detected_game_type = {
            "hk4e_cn": GameType.GenshinImpact,
            "hkrpg_cn": GameType.StarRail
        }[bi]

        if self._game_type is None:
            self._game_type = detected_game_type
        else:
            if self._game_type != detected_game_type:
                warnings.warn(
                    "The program detected a difference between the GameType you entered and the GameType it detected. "
                    "This time, it will use your input as the standard. So the data may be incorrect.”"
                    "Please pay attention to the GameType.",
                    SyntaxWarning
                )

    def get_user_headers(self) -> dict:
        """
        获取该用户的 headers。

        Returns:
            字典格式的该用户的 header。
        """
        return {
            "x-rpc-combo_token": self._combo_token,
            "x-rpc-sys_version": self._sys_version,
            "x-rpc-device_id": self._device_id,
            "x-rpc-device_name": self._device_name,
            "x-rpc-device_model": self._device_model,
            # Client type in headers must be string.
            "x-rpc-client_type": str(self._client_type.value),
            "x-rpc-channel": self._client_type
        }

    @property
    def combo_token(self) -> str:
        return self._combo_token

    @property
    def sys_version(self) -> str:
        return self._sys_version

    @property
    def device_id(self) -> str:
        return self._device_id

    @property
    def device_name(self) -> str:
        return self._device_name

    @property
    def device_model(self) -> str:
        return self._device_model

    @property
    def client_type(self) -> UserClientType:
        return self._client_type

    @property
    def game_type(self) -> GameType:
        return self._game_type

    @property
    def channel(self) -> UserChannel:
        return self._channel



class WebRequestError(RuntimeError):
    def __init__(self, message: str, status_code: int):
        self._message = message
        self._status_code = status_code

    @property
    def message(self):
        return self._message

    @property
    def status_code(self):
        return self._status_code


class APIRequestError(RuntimeError):
    def __init__(self, message: str, ret_code: int):
        self._message = message
        self._ret_code = ret_code

    @property
    def message(self):
        return self._message

    @property
    def ret_code(self):
        return self._ret_code


class ComboTokenInvalidError(RuntimeError):
    def __init__(self, message: str):
        self._message = message

    @property
    def message(self):
        return self._message



from enum import Enum
from typing import Optional, List
from ._types import GameType, NotificationType, NotificationStatus, UserClientType
from ._api import API
from ._user import User
from ._exceptions import WebRequestError, APIRequestError
from ._wallet import WalletData
from ._notification import Notification
import httpx


class ClientStatus(Enum):
    # UNOPENED:
    #   The client has been instantiated, but has not been used to send a web request,
    #   or been opened by entering the context of a `with` block.
    UNOPENED = 0
    # OPENED:
    #   The client has either sent a web request, or is within a `with` block.
    OPENED = 1
    # CLOSED:
    #   The client has either exited the `with` block, or `close()` has been called explicitly.
    CLOSED = 2


class Client:
    """
    米哈云游客户端。
    """

    def __init__(self):
        self._client: httpx.Client = httpx.Client()
        self._status: ClientStatus = ClientStatus.UNOPENED
        # The version will be updated on the first request from the corresponding game.
        self._versions: dict = {
            GameType.GenshinImpact: None,
            GameType.StarRail: None
        }

    def __enter__(self):
        if self._status != ClientStatus.UNOPENED:
            msg = {
                ClientStatus.OPENED: "Cannot open a client instance more than once.",
                ClientStatus.CLOSED: "Cannot reopen a client instance, once it has been closed.",
            }[self._status]
            raise RuntimeError(msg)

        self._status = ClientStatus.OPENED
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._status = ClientStatus.CLOSED
        self._client.close()

    def _update_version(self, game_type: GameType) -> None:
        """
        更新客户端版本。

        Args:
            game_type (GameType): 要更新的客户端类型。
        """
        version_url = API.get_game_version_url(game_type)

        resp = self._client.get(version_url, params={
            "key": API.get_launcher_key(game_type),
            "launcher_id": API.get_launcher_id(game_type)
        }).json()

        self._versions[game_type] = resp["data"]["game"]["latest"]["version"]

    def _get_common_headers(self, game_type: GameType, client_type: UserClientType) -> dict:
        """
        获取指定游戏类型的 headers 常量。

        Args:
            game_type (GameType): 需要获取 headers 的游戏类型。
            client_type (UserClientType): 用户的客户端类型。

        Returns:
            一个字典，包含了指定游戏的 headers。
        """
        return {
            "x-rpc-app_version": self._versions[game_type],
            "x-rpc-app_id": API.get_app_id(game_type, client_type),
            "x-rpc-vendor_id": API.get_vendor_id(game_type, client_type),
            "x-rpc-cg_game_biz": API.get_cg_game_biz(game_type),
            "x-rpc-op_biz": API.get_op_biz(game_type),
            "x-rpc-cps": API.get_cps(game_type, client_type)
        }

    def _user_web_get(self, user: User, url: str, params: Optional[dict] = None) -> httpx.Response:
        """
        附带用户 headers 的 get 类型请求。

        Args:
            user (User): 发起请求的用户。
            url (str): 目标 URL。
            params (Optional[dict]): 可选的参数。

        Returns:
            httpx 库的 Response 类。
        """

        # Check the client state.
        if self._status == ClientStatus.CLOSED:
            raise RuntimeError("Cannot send a request, as the client has been closed.")

        # Check version.
        if self._versions[user.game_type] is None:
            self._update_version(user.game_type)

        # Update client state.
        if self._status != ClientStatus.OPENED:
            self._status = ClientStatus.OPENED

        # Get the special common headers of the game.
        headers: dict = self._get_common_headers(user.game_type, user.client_type)

        user_headers: dict = user.get_user_headers()

        headers.update(user_headers)

        headers["x-rpc-channel"] = API.get_channel_id(user.channel)

        resp = self._client.get(url, headers=headers, params=params)

        if resp.status_code != 200:
            raise WebRequestError(
                f"An error occurred in the network request, status code: {resp.status_code}",
                resp.status_code
            )

        return resp

    def get_wallet_data(self, user: User) -> WalletData:
        """
        获取指定用户的钱包数据。

        Args:
            user (User): 发起请求的用户。

        Returns:
            该用户的钱包数据。
        """
        r = self._user_web_get(user, API.get_wallet_data_url(user.game_type)).json()
        if r['retcode'] != 0:
            raise APIRequestError(r['message'], r['retcode'])
        return WalletData.from_dict(r['data'])

    def get_notifications(
            self,
            user: User,
            *,
            status: Optional[NotificationStatus] = None,
            type_: Optional[NotificationType] = None,
            is_sort: Optional[bool] = True
    ) -> List[Notification]:
        """
        获取指定用户的通知信息。

        Args:
            user (User): 发起请求的用户。
            status (Optional[NotificationStatus]): 筛选指定的通知状态。
            type_ (Optional[NotificationType]): 筛选指定的通知种类。
            is_sort (Optional[bool]): 是否排序。

        Returns:
            一个列表，包含了指定用户的指定通知。
        """

        # Build params.

        params = {
            "is_sort": is_sort
        }

        # Optional params.
        if status:
            params["status"] = status.value
        if type_:
            params["type"] = type_.value

        r = self._user_web_get(
            user,
            API.get_notifications_url(user.game_type),
            params=params
        ).json()

        if r['retcode'] != 0:
            raise APIRequestError(r['message'], r['retcode'])

        notifications = []

        for notification_data in r['data']['list']:
            notifications.append(Notification.from_data_dict(notification_data))

        return notifications

    def get_client_version(self, game_type: GameType) -> str:
        """
        获取指定游戏类型的版本号，若想获取字典类型的所有版本号，请使用 versions 属性。

        Args:
            game_type (GameType): 游戏类型。

        Returns:
            该游戏类型的版本号。
        """
        return self._versions[game_type]

    @property
    def versions(self) -> dict:
        """
        所有游戏类型的版本号字典，若只想获取指定游戏类型的版本号，请使用 get_client_version() 方法。
        """
        return self._versions



# Client
from ._client import Client
from ._types import GameType

# User
from ._user import User
from ._types import UserClientType, UserChannel

# Notification
from ._types import NotificationType, NotificationStatus



from ._types import GameType, UserChannel, UserClientType


class API:
    @staticmethod
    def get_launcher_key(game_type: GameType) -> str:
        return {
            GameType.GenshinImpact: "eYd89JmJ",
            GameType.StarRail: "6KcVuOkbcqjJomjZ"
        }[game_type]

    @staticmethod
    def get_launcher_id(game_type: GameType) -> str:
        return {
            GameType.GenshinImpact: "18",
            GameType.StarRail: "33"
        }[game_type]

    @staticmethod
    def get_game_version_url(game_type: GameType) -> str:
        return {
            GameType.GenshinImpact: "https://sdk-static.mihoyo.com/hk4e_cn/mdk/launcher/api/resource",
            GameType.StarRail: "https://api-launcher.mihoyo.com/hkrpg_cn/mdk/launcher/api/resource"
        }[game_type]

    @staticmethod
    def get_app_id(game_type: GameType, client_type: UserClientType) -> str:
        return {
            UserClientType.Android: {
                GameType.GenshinImpact: "1953439974",
                GameType.StarRail: "1953445976",
            },
            UserClientType.PCWeb: {
                GameType.GenshinImpact: "4",
                GameType.StarRail: "8"
            }
        }[client_type][game_type]

    @staticmethod
    def get_vendor_id(game_type: GameType, client_type: UserClientType) -> str:
        return {
            UserClientType.Android: {
                GameType.GenshinImpact: "1",
                GameType.StarRail: "2",
            },
            UserClientType.PCWeb: {
                GameType.GenshinImpact: "2",
                GameType.StarRail: "2"
            }
        }[client_type][game_type]

    @staticmethod
    def get_cg_game_biz(game_type: GameType) -> str:
        return {
            GameType.GenshinImpact: "hk4e_cn",
            GameType.StarRail: "hkrpg_cn",
        }[game_type]

    @staticmethod
    def get_op_biz(game_type: GameType) -> str:
        return {
            GameType.GenshinImpact: "clgm_cn",
            GameType.StarRail: "clgm_hkrpg-cn",
        }[game_type]

    @staticmethod
    def get_cps(game_type: GameType, client_type: UserClientType) -> str:
        return {
            UserClientType.Android: {
                GameType.GenshinImpact: "mihoyo",
                GameType.StarRail: "mihoyo",
            },
            UserClientType.PCWeb: {
                GameType.GenshinImpact: "pc_default",
                GameType.StarRail: "pc_official"
            }
        }[client_type][game_type]

    @staticmethod
    def get_channel_id(user_channel: UserChannel) -> str:
        return {
            UserChannel.Official: "mihoyo"
        }[user_channel]

    @staticmethod
    def get_wallet_data_url(game_type: GameType) -> str:
        return {
            GameType.GenshinImpact: "https://api-cloudgame.mihoyo.com/hk4e_cg_cn/wallet/wallet/get",
            GameType.StarRail: "https://cg-hkrpg-api.mihoyo.com/hkrpg_cn/cg/wallet/wallet/get"
        }[game_type]

    @staticmethod
    def get_notifications_url(game_type: GameType) -> str:
        return {
            GameType.GenshinImpact: "https://api-cloudgame.mihoyo.com/hk4e_cg_cn/gamer/api/listNotifications",
            GameType.StarRail: "https://cg-hkrpg-api.mihoyo.com/hkrpg_cn/cg/gamer/api/listNotifications"
        }[game_type]



from enum import Enum
from typing import TypeVar


class GameType(Enum):
    """
    游戏类型。

    Attributes:
        GenshinImpact: 云·原神。
        StarRail: 云·星穹铁道。
    """
    GenshinImpact = 0
    StarRail = 1


class UserClientType(Enum):
    """
    客户端类型。

    Attributes:
        Android: 安卓。
        PCWeb: PC 网页版。
    """
    Android = 2
    PCWeb = 16


class UserChannel(Enum):
    """
    游戏渠道。

    Attributes:
        Official: 官方服。
    """
    Official = 0


T = TypeVar("T", bound="NotificationStatus")


class NotificationStatus(Enum):
    """
    通知状态。

    Attributes:
        Read: 已读。
        Unread: 未读。
        Undefined: 未定义。
    """
    Read = "NotificationStatusRead"
    Unread = "NotificationStatusUnread"

    Undefined = "NotificationStatusUndefined"

    @classmethod
    def get_status_by_name(cls, status: str) -> T:
        """
        从字符串获取枚举成员。

        Args:
            status (str): 成员字符串。

        Returns:
            枚举成员。
        """
        for member in cls:
            if member.value == status:
                return member
        return cls.Undefined


V = TypeVar("V", bound="NotificationType")


class NotificationType(Enum):
    """
    通知种类。

    Attributes:
        Popup: 弹窗通知。
        Undefined: 未定义。
    """
    Popup = "NotificationTypePopup"

    Undefined = "NotificationTypeUndefined"

    @classmethod
    def get_type_by_name(cls, type_: str) -> V:
        """
        从字符串获取枚举成员。

        Args:
            type_ (str): 成员字符串。

        Returns:
            枚举成员。
        """
        for member in cls:
            if member.value == type_:
                return member
        return cls.Undefined



from dataclasses import dataclass
from dataclasses_json import dataclass_json


@dataclass_json
@dataclass(frozen=True)
class CoinData:
    """
    用户的 原点 / 星云币 时长数据。

    Attributes:
        coin_num (int): 原点 / 星云币 数。
        free_coin_num (int): 免费 原点 / 星云币 数。
        coin_limit (int): 原点 / 星云币 的数量上限。
        exchange (int): 与原点时长的汇率。通常来说，10 原点 / 星云币 = 1 游戏时长。
    """
    coin_num: int
    free_coin_num: int
    coin_limit: int
    exchange: int


@dataclass_json
@dataclass(frozen=True)
class FreeTimeData:
    """
    用户的免费时长数据。

    Attributes:
        send_freetime (int): 若该用户是本日第一次登录，那么此处的值为每日赠送的免费时长 (min)。反之，此处恒为 0。
        free_time (int): 总免费时长 (min)。
        free_time_limit (int): 免费时长上限 (min)。
        over_freetime (int): 该用户是本日第一次登录且赠送的时长有一部分超出了免费时长上限，那么此处的值为溢出的免费时长 (min)。
    """
    send_freetime: int
    free_time: int
    free_time_limit: int
    over_freetime: int


@dataclass_json
@dataclass(frozen=True)
class StatusData:
    """
    未知数据
    """
    status: int
    msg: str
    total_time_status: int
    status_new: int


@dataclass_json
@dataclass(frozen=True)
class StatData:
    """
    未知数据
    """
    vip_point: str


@dataclass_json
@dataclass(frozen=True)
class PlayCardData:
    """
    用户的畅玩卡数据

    Attributes:
        expire (str): 畅玩卡过期时间
        msg (str): 畅玩卡信息
        short_msg (str): 畅玩卡短信息 / 状态
        play_card_limit (str): 未知
        remaining_sec (str): 未知
    """
    expire: str
    msg: str
    short_msg: str
    play_card_limit: str
    remaining_sec: str


@dataclass_json
@dataclass(frozen=True)
class WalletData:
    """
    用户的钱包数据。

    Attributes:
        coin (CoinData): 付费时长数据。
        free_time (FreeTimeData): 免费时长数据。
        status (StatusData): 未知。
        stat (StatData): 未知。
        play_card (PlayCardData): 畅玩卡数据。
        total_time (int): 总时长。
    """

    coin: CoinData
    free_time: FreeTimeData
    status: StatusData
    stat: StatData
    play_card: PlayCardData
    total_time: int

    def is_sign_in(self) -> bool:
        """
        判断本次行为是否为签到操作 (本日第一次登录)。

        Returns:
            若为 True，则本次行为是签到操作。
        """
        return self.free_time.send_freetime != 0



from typing import TypeVar
from ._types import NotificationType, NotificationStatus


T = TypeVar("T", bound="Notification")


class Notification:
    """
    通知类。
    """
    def __init__(
            self,
            id_: str,
            status: NotificationStatus,
            type_: NotificationType,
            priority: int,
            source: str,
            desc: str,
            msg: str,
            created_at: str
    ):
        self._id = id_
        self._status = status
        self._type = type_
        self._priority = priority
        self._source = source
        self._desc = desc
        self._msg = msg
        self._created_at = created_at

    def __repr__(self) -> str:
        return (f"Notification(id={self._id}, status={self._status}, type={self._type}, priority={self._priority}, "
                f"source={self._source}, desc={self._desc}, msg={self.msg}, created_at={self._created_at})")

    @classmethod
    def from_data_dict(cls, data: dict) -> T:
        """
        从特定的数据结构生成 Notification。

        Args:
            data (dict): 消息数据。

        Returns:
            包装后的消息数据。
        """
        return cls(
            id_=data['id'],
            status=NotificationStatus.get_status_by_name(data['status']),
            type_=NotificationType.get_type_by_name(data['type']),
            priority=data['priority'],
            source=data['source'],
            desc=data['desc'],
            msg=data['msg'],
            created_at=data['created_at']
        )

    @property
    def id(self) -> str:
        """
        通知 ID。

        Returns:
            该通知的 ID。
        """
        return self._id

    @property
    def status(self) -> NotificationStatus:
        """
        通知状态。

        Returns:
            该通知的通知状态。
        """
        return self._status

    @property
    def type(self) -> NotificationType:
        """
        通知种类。

        Returns:
            该通知的通知种类。
        """
        return self._type

    @property
    def priority(self) -> int:
        """
        作用未知，根据名称推测是通知的优先级。

        Returns:
            该通知的优先级。
        """
        return self._priority

    @property
    def source(self) -> str:
        """
        作用未知。

        Returns:
            一个字符串，作用未知。
        """
        return self._source

    @property
    def desc(self) -> str:
        """
        作用未知，根据名称推测是通知的描述。

        Returns:
            该通知的描述。
        """
        return self._desc

    @property
    def msg(self) -> str:
        """
        通知内容。

        Returns:
            一个字符串，包含了 json 文本格式的该通知的内容。
        """
        return self._msg

    @property
    def create_at(self):
        """
        通知创建时间。

        Returns:
            一个字符串，是秒级的时间戳 (10位)，描述了该通知何时被创建。
        """
        return self._created_at
